@@ -25,10 +25,24 @@
 #include <unistd.h>
 #include <time.h>
 #include <linux/fs.h>
-#include <corosync/engine/logsys.h>
+#include <liblogthread.h>
+#include <zlib.h>
+#include "iostate.h"
 
 static int diskRawRead(target_info_t *disk, char *buf, int len);
-uint32_t clu_crc32(const char *data, size_t count);
+
+/**
+ * Calculate CRC32 of a data set.
+ *
+ * @param data		Data set for building CRC32
+ * @param count		Size of data set, in bytes.
+ * @return 		CRC32 of data set.
+ */
+static uint32_t clu_crc32(const char *data, size_t count)
+{
+	return (uint32_t)crc32(0L, (const Bytef *)data, (uInt)count);
+}
+
 
 /**
  * Swap the bytes of a shared header so that it's always in big-endian form
@@ -40,15 +54,15 @@ static void
 header_encode(shared_header_t *hdr)
 {
 	/* sanity check - LE machine -> already encoded. */
-	if (hdr->h_magic == be_swap32(SHARED_HEADER_MAGIC))
+	if (hdr->h_magic == le_swap32(SHARED_HEADER_MAGIC))
 		return;
 
-	swab32(hdr->h_magic);
-	swab32(hdr->h_hcrc);
-	swab32(hdr->h_dcrc);
-	swab32(hdr->h_length);
-	swab64(hdr->h_view);
-	swab64(hdr->h_timestamp);
+	hdr->h_magic = le_swap32(hdr->h_magic);
+	hdr->h_hcrc = le_swap32(hdr->h_hcrc);
+	hdr->h_dcrc = le_swap32(hdr->h_dcrc);
+	hdr->h_length = le_swap32(hdr->h_length);
+	hdr->h_view = le_swap64(hdr->h_view);
+	hdr->h_timestamp = le_swap64(hdr->h_timestamp);
 }
 
 
@@ -65,12 +79,12 @@ header_decode(shared_header_t *hdr)
 	if (hdr->h_magic == SHARED_HEADER_MAGIC)
 		return;
 
-	swab32(hdr->h_magic);
-	swab32(hdr->h_hcrc);
-	swab32(hdr->h_dcrc);
-	swab32(hdr->h_length);
-	swab64(hdr->h_view);
-	swab64(hdr->h_timestamp);
+	hdr->h_magic = le_swap32(hdr->h_magic);
+	hdr->h_hcrc = le_swap32(hdr->h_hcrc);
+	hdr->h_dcrc = le_swap32(hdr->h_dcrc);
+	hdr->h_length = le_swap32(hdr->h_length);
+	hdr->h_view = le_swap64(hdr->h_view);
+	hdr->h_timestamp = le_swap64(hdr->h_timestamp);
 }
 
 
@@ -100,21 +114,21 @@ header_generate(shared_header_t *hdr, const char *data, size_t count)
 		hdr->h_length = (uint32_t)count;
 
 		if (hdr->h_dcrc == 0) {
-			log_printf(LOG_ERR, "Invalid CRC32 generated on data!\n");
+			logt_print(LOG_ERR, "Invalid CRC32 generated on data!\n");
 			return -1;
 		}
 	}
 
 	hdr->h_timestamp = (uint64_t)time(NULL);
+	header_encode(hdr);
+	hdr->h_hcrc = 0;
+	hdr->h_hcrc = le_swap32(clu_crc32((char *)hdr, sizeof(*hdr)));
 
-	hdr->h_hcrc = clu_crc32((char *)hdr, sizeof(*hdr));
 	if (hdr->h_hcrc == 0) {
-		log_printf(LOG_ERR, "Invalid CRC32 generated on header!\n");
+		logt_print(LOG_ERR, "Invalid CRC32 generated on header!\n");
 		return -1;
 	}
 
-	header_encode(hdr);
-
 	return 0;
 }
 
@@ -135,27 +149,32 @@ header_verify(shared_header_t *hdr, const char *data, size_t count)
 	uint32_t crc;
 	uint32_t bkupcrc;
 
-	header_decode(hdr);
 	/*
 	 * verify the header's CRC32.  Ok, we know it's overkill taking
 	 * the CRC32 of a friggin' 16-byte (12 bytes, really) structure,
 	 * but why not?
 	 */
 	bkupcrc = hdr->h_hcrc;
+
+	/* BUG: Headers are stored in little-endian form */
+	bkupcrc = le_swap32(hdr->h_hcrc); 
+
 	hdr->h_hcrc = 0;
 	crc = clu_crc32((char *)hdr, sizeof(*hdr));
 	hdr->h_hcrc = bkupcrc;
 	if (bkupcrc != crc) {
-		log_printf(LOG_DEBUG, "Header CRC32 mismatch; Exp: 0x%08x "
+		logt_print(LOG_DEBUG, "Header CRC32 mismatch; Exp: 0x%08x "
 			"Got: 0x%08x\n", bkupcrc, crc);
 		return -1;
 	}
 
+	header_decode(hdr);
+
 	/*
 	 * Verify the magic number.
 	 */
 	if (hdr->h_magic != SHARED_HEADER_MAGIC) {
-		log_printf(LOG_DEBUG, "Magic mismatch; Exp: 0x%08x "
+		logt_print(LOG_DEBUG, "Magic mismatch; Exp: 0x%08x "
 			"Got: 0x%08x\n", SHARED_HEADER_MAGIC, hdr->h_magic);
 		return -1;
 	}
@@ -171,7 +190,7 @@ header_verify(shared_header_t *hdr, const char *data, size_t count)
 			hdr->h_length : count);
 
 	if (hdr->h_dcrc != crc) {
-		log_printf(LOG_DEBUG, "Data CRC32 mismatch; Exp: 0x%08x "
+		logt_print(LOG_DEBUG, "Data CRC32 mismatch; Exp: 0x%08x "
 			"Got: 0x%08x\n", hdr->h_dcrc, crc);
 		return -1;
 	}
@@ -202,7 +221,8 @@ qdisk_open(char *name, target_info_t *disk)
 
 	ret = ioctl(disk->d_fd, BLKSSZGET, &ssz);
 	if (ret < 0) {
-		log_printf(LOG_ERR, "qdisk_open: ioctl(BLKSSZGET)");
+		logt_print(LOG_ERR, "qdisk_open: ioctl(BLKSSZGET)");
+		close(disk->d_fd);
 		return -1;
 	}
 
@@ -210,29 +230,33 @@ qdisk_open(char *name, target_info_t *disk)
 	disk->d_pagesz = sysconf(_SC_PAGESIZE);
 
 	/* Check to verify that the partition is large enough.*/
+	io_state(STATE_LSEEK);
 	ret = lseek(disk->d_fd, END_OF_DISK(disk->d_blksz), SEEK_SET);
+	io_state(STATE_NONE);
 	if (ret < 0) {
-		log_printf(LOG_DEBUG, "open_partition: seek");
+		logt_print(LOG_DEBUG, "open_partition: seek");
+		close(disk->d_fd);
 		return -1;
 	}
 
 	if (ret < END_OF_DISK(disk->d_blksz)) {
-		log_printf(LOG_ERR, "Partition %s too small\n", name);
+		logt_print(LOG_ERR, "Partition %s too small\n", name);
 		errno = EINVAL;
+		close(disk->d_fd);
 		return -1;
 	}
 
 	/* Set close-on-exec bit */
         ret = fcntl(disk->d_fd, F_GETFD, 0);
         if (ret < 0) {
-		log_printf(LOG_ERR, "open_partition: fcntl(F_GETFD)");
+		logt_print(LOG_ERR, "open_partition: fcntl(F_GETFD)");
                 close(disk->d_fd);
                 return -1;
         }
 
         ret |= FD_CLOEXEC;
         if (fcntl(disk->d_fd, F_SETFD, ret) < 0) {
-		log_printf(LOG_ERR, "open_partition: fcntl(F_SETFD)");
+		logt_print(LOG_ERR, "open_partition: fcntl(F_SETFD)");
                 close(disk->d_fd);
                 return -1;
         }
@@ -276,7 +300,7 @@ qdisk_validate(char *name)
 	stat_ptr = &stat_st;
 
 	if (stat(name, stat_ptr) < 0) {
-		log_printf(LOG_ERR, "stat");
+		logt_print(LOG_ERR, "stat");
 		return -1;
 	}
 	/*
@@ -287,7 +311,7 @@ qdisk_validate(char *name)
 		errno = EINVAL;
 		return -1;
 */
-		log_printf(LOG_WARNING, "Warning: %s is not a block device\n",
+		logt_print(LOG_WARNING, "Warning: %s is not a block device\n",
 		        name);
 	}
 
@@ -295,7 +319,7 @@ qdisk_validate(char *name)
 	 * Verify read/write permission.
 	 */
 	if (qdisk_open(name, &disk) < 0) {
-		log_printf(LOG_DEBUG, "%s: open of %s for RDWR failed: %s\n",
+		logt_print(LOG_DEBUG, "%s: open of %s for RDWR failed: %s\n",
 			__FUNCTION__, name, strerror(errno));
 		return -1;
 	}
@@ -310,11 +334,12 @@ diskRawReadShadow(target_info_t *disk, off_t readOffset, char *buf, int len)
 	int ret;
 	shared_header_t *hdrp;
 	char *data;
-	int datalen;
 
+	io_state(STATE_LSEEK);
 	ret = lseek(disk->d_fd, readOffset, SEEK_SET);
+	io_state(STATE_NONE);
 	if (ret != readOffset) {
-		log_printf(LOG_DEBUG,
+		logt_print(LOG_DEBUG,
 		       "diskRawReadShadow: can't seek to offset %d.\n",
 		       (int) readOffset);
 		errno = ENODATA;
@@ -323,7 +348,7 @@ diskRawReadShadow(target_info_t *disk, off_t readOffset, char *buf, int len)
 
 	ret = diskRawRead(disk, buf, len);
 	if (ret != len) {
-		log_printf(LOG_DEBUG, "diskRawReadShadow: aligned read "
+		logt_print(LOG_DEBUG, "diskRawReadShadow: aligned read "
 		       "returned %d, not %d.\n", ret, len);
 		errno = ENODATA;
 		return -1;
@@ -332,11 +357,9 @@ diskRawReadShadow(target_info_t *disk, off_t readOffset, char *buf, int len)
 	/* Decode the header portion so we can run a checksum on it. */
 	hdrp = (shared_header_t *)buf;
 	data = (char *)buf + sizeof(*hdrp);
-	swab_shared_header_t(hdrp);
-	datalen = hdrp->h_length;
 
 	if (header_verify(hdrp, data, len)) {
-		log_printf(LOG_DEBUG, "diskRawReadShadow: bad CRC32, "
+		logt_print(LOG_DEBUG, "diskRawReadShadow: bad CRC32, "
 		       "offset = %d len = %d\n",
 		       (int) readOffset, len);
 		errno = EPROTO;
@@ -371,11 +394,14 @@ diskRawRead(target_info_t *disk, char *buf, int len)
 	if (bounceNeeded == 0) {
 		/* Already aligned and even multiple of 512, no bounceio
 		 * required. */
-		return (read(disk->d_fd, buf, len));
+		io_state(STATE_READ);
+		readret = read(disk->d_fd, buf, len);
+		io_state(STATE_NONE);
+		return readret;
 	}
 
 	if (len > disk->d_blksz) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 			"diskRawRead: not setup for reads larger than %d.\n",
 		       (int)disk->d_blksz);
 		return (-1);
@@ -400,7 +426,9 @@ diskRawRead(target_info_t *disk, char *buf, int len)
 		return -1;
 	}
 
+	io_state(STATE_READ);
 	readret = read(disk->d_fd, alignedBuf, readlen);
+	io_state(STATE_NONE);
 	if (readret > 0) {
 		if (readret > len) {
 			memcpy(alignedBuf, buf, len);
@@ -412,7 +440,7 @@ diskRawRead(target_info_t *disk, char *buf, int len)
 
 	free(alignedBuf);
 	if (readret != len) {
-		log_printf(LOG_ERR, "diskRawRead: read err, len=%d, readret=%d\n",
+		logt_print(LOG_ERR, "diskRawRead: read err, len=%d, readret=%d\n",
 			len, readret);
 	}
 
@@ -443,11 +471,14 @@ diskRawWrite(target_info_t *disk, char *buf, int len)
 	if (bounceNeeded == 0) {
 		/* Already aligned and even multiple of 512, no bounceio
 		 * required. */
-		return (write(disk->d_fd, buf, len));
+		io_state(STATE_WRITE);
+		ret = write(disk->d_fd, buf, len);
+		io_state(STATE_NONE);
+		return ret;
 	}
 
 	if (len > disk->d_blksz) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 			"diskRawRead: not setup for reads larger than %d.\n",
 		       (int)disk->d_blksz);
 		return (-1);
@@ -473,21 +504,23 @@ diskRawWrite(target_info_t *disk, char *buf, int len)
 	}
 
 	if (len > disk->d_blksz) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "diskRawWrite: not setup for larger than %d.\n",
 		       (int)disk->d_blksz);
 		return (-1);
 	}
 
 	memcpy(buf, alignedBuf, len);
+	io_state(STATE_WRITE);
 	ret = write(disk->d_fd, alignedBuf, writelen);
+	io_state(STATE_NONE);
 	if (ret > len) {
 		ret = len;
 	}
 
 	free(alignedBuf);
 	if (ret != len) {
-		log_printf(LOG_ERR, "diskRawWrite: write err, len=%d, ret=%dn",
+		logt_print(LOG_ERR, "diskRawWrite: write err, len=%d, ret=%dn",
 		       len, ret);
 	}
 
@@ -502,15 +535,17 @@ diskRawWriteShadow(target_info_t *disk, __off64_t writeOffset, char *buf, int le
 	ssize_t retval_write;
 
 	if ((writeOffset < 0) || (len < 0)) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "diskRawWriteShadow: writeOffset=%08x, "
 		       "len=%08x.\n", (int)writeOffset, len);
 		return (-1);
 	}
 
+	io_state(STATE_LSEEK);
 	retval_seek = lseek(disk->d_fd, writeOffset, SEEK_SET);
+	io_state(STATE_NONE);
 	if (retval_seek != writeOffset) {
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "diskRawWriteShadow: can't seek to offset %d\n",
 		       (int) writeOffset);
 		return (-1);
@@ -519,10 +554,10 @@ diskRawWriteShadow(target_info_t *disk, __off64_t writeOffset, char *buf, int le
 	retval_write = diskRawWrite(disk, buf, len);
 	if (retval_write != len) {
 		if (retval_write == -1) {
-			log_printf(LOG_ERR, "%s: %s\n", __FUNCTION__,
+			logt_print(LOG_ERR, "%s: %s\n", __FUNCTION__,
 			       strerror(errno));
 		}
-		log_printf(LOG_ERR,
+		logt_print(LOG_ERR,
 		       "diskRawWriteShadow: aligned write returned %d"
 		       ", not %d\n", (int)retval_write, (int)len);
 		return (-1);
@@ -533,13 +568,14 @@ diskRawWriteShadow(target_info_t *disk, __off64_t writeOffset, char *buf, int le
 
 
 int
-qdisk_read(target_info_t *disk, __off64_t offset, void *buf, int count)
+qdisk_read(target_info_t *disk, __off64_t offset, void *bufin, int count)
 {
 	shared_header_t *hdrp;
 	void *ptr;
 	char *data;
 	size_t total;
 	int rv;
+	char *buf = (char *)bufin;
 
 	/*
 	 * Calculate the total length of the buffer, including the header.
@@ -595,7 +631,7 @@ qdisk_write(target_info_t *disk, __off64_t offset, const void *buf, int count)
 
 	maxsize = psz - (sizeof(shared_header_t));
 	if (count >= (maxsize + sizeof(shared_header_t))) {
-		log_printf(LOG_ERR, "error: count %d >= (%d + %d)\n", (int)count,
+		logt_print(LOG_ERR, "error: count %d >= (%d + %d)\n", (int)count,
 		       (int)maxsize, (int)sizeof(shared_header_t));
 		errno = ENOSPC;
 		return -1;
@@ -615,7 +651,7 @@ qdisk_write(target_info_t *disk, __off64_t offset, const void *buf, int count)
 	ptr = NULL;
 	rv = posix_memalign((void **)&ptr, disk->d_pagesz, total);
 	if (rv < 0) {
-		log_printf(LOG_ERR, "posix_memalign");
+		logt_print(LOG_ERR, "posix_memalign");
 		return -1;
 	}
 
@@ -630,7 +666,6 @@ qdisk_write(target_info_t *disk, __off64_t offset, const void *buf, int count)
 		free((char *)hdrp);
 		return -1;
 	}
-	swab_shared_header_t(hdrp);
 
 	/* 
 	 * Locking must be performed elsewhere.  We make no assumptions
@@ -640,7 +675,7 @@ qdisk_write(target_info_t *disk, __off64_t offset, const void *buf, int count)
 		rv = diskRawWriteShadow(disk, offset, (char *)hdrp, psz);
 
 	if (rv == -1)
-		log_printf(LOG_ERR, "diskRawWriteShadow");
+		logt_print(LOG_ERR, "diskRawWriteShadow");
 	
 	free((char *)hdrp);
 	if (rv == -1)
@@ -666,7 +701,7 @@ header_init(target_info_t *disk, char *label)
 	}
 
 	if (gethostname(qh.qh_updatehost, sizeof(qh.qh_updatehost)) < 0) {
-		log_printf(LOG_ERR, "gethostname");
+		logt_print(LOG_ERR, "gethostname");
 		return -1;
 	}
 
@@ -675,7 +710,7 @@ header_init(target_info_t *disk, char *label)
 
 	qh.qh_version = VERSION_MAGIC_V2;
 	if ((qh.qh_timestamp = (uint64_t)time(NULL)) <= 0) {
-		log_printf(LOG_ERR, "time");
+		logt_print(LOG_ERR, "time");
 		return -1;
 	}
 
@@ -702,13 +737,13 @@ qdisk_init(char *partname, char *label)
 
 	ret = qdisk_validate(partname);
 	if (ret < 0) {
-		log_printf(LOG_DEBUG, "qdisk_verify");
+		logt_print(LOG_DEBUG, "qdisk_verify");
 		return -1;
 	}
 
 	ret = qdisk_open(partname, &disk);
 	if (ret < 0) {
-		log_printf(LOG_ERR, "qdisk_open");
+		logt_print(LOG_ERR, "qdisk_open");
 		return -1;
 	}
 
