@@ -41,7 +41,7 @@
  */
 
 static void
-print_array(char *title, char *buf, int count, int console)
+print_array(const char *title, char *buf, int count, int console)
 {
 	int x;
 
@@ -349,20 +349,20 @@ gfs_jindex_print(struct gfs_jindex *ji, int console)
  */
 
 void
-gfs_rindex_in(struct gfs_rindex *rindex, char *buf)
+gfs_rindex_in(struct gfs_rindex *grindex, char *buf)
 {
 	struct gfs_rindex *str = (struct gfs_rindex *) buf;
 
-	CPIN_64(rindex, str, ri_addr);
-	CPIN_32(rindex, str, ri_length);
-	CPIN_32(rindex, str, ri_pad);
+	CPIN_64(grindex, str, ri_addr);
+	CPIN_32(grindex, str, ri_length);
+	CPIN_32(grindex, str, ri_pad);
 
-	CPIN_64(rindex, str, ri_data1);
-	CPIN_32(rindex, str, ri_data);
+	CPIN_64(grindex, str, ri_data1);
+	CPIN_32(grindex, str, ri_data);
 
-	CPIN_32(rindex, str, ri_bitbytes);
+	CPIN_32(grindex, str, ri_bitbytes);
 
-	CPIN_08(rindex, str, ri_reserved, 64);
+	CPIN_08(grindex, str, ri_reserved, 64);
 
 }
 
@@ -374,20 +374,20 @@ gfs_rindex_in(struct gfs_rindex *rindex, char *buf)
  */
 
 void
-gfs_rindex_out(struct gfs_rindex *rindex, char *buf)
+gfs_rindex_out(struct gfs_rindex *grindex, char *buf)
 {
 	struct gfs_rindex *str = (struct gfs_rindex *) buf;
 
-	CPOUT_64(rindex, str, ri_addr);
-	CPOUT_32(rindex, str, ri_length);
-	CPOUT_32(rindex, str, ri_pad);
+	CPOUT_64(grindex, str, ri_addr);
+	CPOUT_32(grindex, str, ri_length);
+	CPOUT_32(grindex, str, ri_pad);
 
-	CPOUT_64(rindex, str, ri_data1);
-	CPOUT_32(rindex, str, ri_data);
+	CPOUT_64(grindex, str, ri_data1);
+	CPOUT_32(grindex, str, ri_data);
 
-	CPOUT_32(rindex, str, ri_bitbytes);
+	CPOUT_32(grindex, str, ri_bitbytes);
 
-	CPOUT_08(rindex, str, ri_reserved, 64);
+	CPOUT_08(grindex, str, ri_reserved, 64);
 
 }
 
