@@ -7,6 +7,7 @@
 
 #define LOCKFILE_NAME	"/var/run/gfs_controld.pid"
 #define CLIENT_NALLOC   32
+#define UEVENT_BUF_SIZE 4096
 
 static int client_maxi;
 static int client_size;
@@ -22,14 +23,14 @@ struct client {
 	struct mountgroup *mg;
 };
 
-static void do_leave(char *table, int mnterr);
+static void do_withdraw(char *name);
 
 int do_read(int fd, void *buf, size_t count)
 {
 	int rv, off = 0;
 
 	while (off < count) {
-		rv = read(fd, buf + off, count - off);
+		rv = read(fd, (char *)buf + off, count - off);
 		if (rv == 0)
 			return -1;
 		if (rv == -1 && errno == EINTR)
@@ -46,7 +47,7 @@ int do_write(int fd, void *buf, size_t count)
 	int rv, off = 0;
 
  retry:
-	rv = write(fd, buf + off, count);
+	rv = write(fd, (char *)buf + off, count);
 	if (rv == -1 && errno == EINTR)
 		goto retry;
 	if (rv < 0) {
@@ -198,135 +199,179 @@ struct mountgroup *find_mg_id(uint32_t id)
 	return NULL;
 }
 
-#define MAXARGS 8
-
-static char *get_args(char *buf, int *argc, char **argv, char sep, int want)
-{
-	char *p = buf, *rp = NULL;
-	int i;
-
-	argv[0] = p;
-
-	for (i = 1; i < MAXARGS; i++) {
-		p = strchr(buf, sep);
-		if (!p)
-			break;
-		*p = '\0';
-
-		if (want == i) {
-			rp = p + 1;
-			break;
-		}
-
-		argv[i] = p + 1;
-		buf = p + 1;
-	}
-	*argc = i;
-
-	/* we ended by hitting \0, return the point following that */
-	if (!rp)
-		rp = strchr(buf, '\0') + 1;
-
-	return rp;
-}
-
-static void ping_kernel_mount(char *table)
+static void ping_kernel_mount(char *name)
 {
 	struct mountgroup *mg;
-	char *name = strstr(table, ":") + 1;
 	int rv, val;
 
 	mg = find_mg(name);
 	if (!mg)
 		return;
 
-	rv = read_sysfs_int(mg, "id", &val);
+	rv = read_sysfs_int(mg, "block", &val);
 
 	log_group(mg, "ping_kernel_mount %d", rv);
 }
 
+enum {
+	Env_ACTION = 0,
+	Env_SUBSYSTEM,
+	Env_LOCKPROTO,
+	Env_LOCKTABLE,
+	Env_DEVPATH,
+	Env_RECOVERY,
+	Env_FIRSTMOUNT,
+	Env_JID,
+	Env_Last, /* Flag for end of vars */
+};
+
+static const char *uevent_vars[] = {
+	[Env_ACTION]		= "ACTION=",
+	[Env_SUBSYSTEM]		= "SUBSYSTEM=",
+	[Env_LOCKPROTO]		= "LOCKPROTO=",
+	[Env_LOCKTABLE]		= "LOCKTABLE=",
+	[Env_DEVPATH]		= "DEVPATH=",
+	[Env_RECOVERY]		= "RECOVERY=",
+	[Env_FIRSTMOUNT]	= "FIRSTMOUNT=",
+	[Env_JID]		= "JID=",
+};
+
+/*
+ * Parses a uevent message for the interesting bits. It requires a list
+ * of variables to look for, and an equally long list of pointers into
+ * which to write the results.
+ */
+static void decode_uevent(const char *buf, unsigned len, const char *vars[],
+			  unsigned nvars, const char *vals[])
+{
+	const char *ptr;
+	unsigned int i;
+	int slen, vlen;
+
+	memset(vals, 0, sizeof(const char *) * nvars);
+
+	while (len > 0) {
+		ptr = buf;
+		slen = strlen(ptr);
+		buf += slen;
+		len -= slen;
+		buf++;
+		len--;
+
+		for (i = 0; i < nvars; i++) {
+			vlen = strlen(vars[i]);
+			if (vlen > slen)
+				continue;
+			if (memcmp(vars[i], ptr, vlen) != 0)
+				continue;
+			vals[i] = ptr + vlen;
+			break;
+		}
+	}
+}
+
+static char *uevent_fsname(const char *vars[])
+{
+	char *name = NULL;
+
+	if (vars[Env_LOCKTABLE])
+		name = strchr(vars[Env_LOCKTABLE], ':');
+
+	/* When all kernels are converted, we can dispose with the following
+	 * grotty bit. This is for backward compatibility only.
+	 */
+	if (!name && vars[Env_DEVPATH]) {
+		name = strchr(vars[Env_DEVPATH], ':');
+		if (name) {
+			char *end = strstr(name, "/lock_module");
+			if (end)
+				*end = 0;
+		}
+	}
+	return (name && name[0]) ? name + 1 : NULL;
+}
+
 static void process_uevent(int ci)
 {
-	char buf[MAXLINE];
-	char *argv[MAXARGS], *act, *sys;
-	int rv, argc = 0;
-	int lock_module = 0;
-
-	memset(buf, 0, sizeof(buf));
-	memset(argv, 0, sizeof(char *) * MAXARGS);
+	char buf[UEVENT_BUF_SIZE];
+	const char *uevent_vals[Env_Last];
+	char *fsname;
+	int rv;
 
  retry_recv:
 	rv = recv(client[ci].fd, &buf, sizeof(buf), 0);
-	if (rv == -1 && rv == EINTR)
-		goto retry_recv;
-	if (rv == -1 && rv == EAGAIN)
-		return;
 	if (rv < 0) {
-		log_error("uevent recv error %d errno %d", rv, errno);
+		if (errno == EINTR)
+			goto retry_recv;
+		if (errno != EAGAIN)
+			log_error("uevent recv error %d errno %d", rv, errno);
+		return;
+	}
+	buf[rv] = 0;
+
+	decode_uevent(buf, rv, uevent_vars, Env_Last, uevent_vals);
+
+	if (!uevent_vals[Env_DEVPATH] ||
+	    !uevent_vals[Env_ACTION] ||
+	    !uevent_vals[Env_SUBSYSTEM])
+		return;
+
+	if (!strstr(uevent_vals[Env_DEVPATH], "/fs/gfs"))
+		return;
+
+	log_debug("uevent %s %s %s",
+		  uevent_vals[Env_ACTION],
+		  uevent_vals[Env_SUBSYSTEM],
+		  uevent_vals[Env_DEVPATH]);
+
+	fsname = uevent_fsname(uevent_vals);
+	if (!fsname) {
+		log_error("no fsname uevent %s %s %s",
+		  	  uevent_vals[Env_ACTION],
+		  	  uevent_vals[Env_SUBSYSTEM],
+		  	  uevent_vals[Env_DEVPATH]);
 		return;
 	}
 
-	/* first we get the uevent for removing lock module kobject:
-	     "remove@/fs/gfs/bull:x/lock_module"
-	   second is the uevent for removing gfs kobject:
-	     "remove@/fs/gfs/bull:x"
-	*/
-
-	if (!strstr(buf, "gfs"))
-		return;
-
-	/* if an fs is named "gfs", it results in dlm uevents
-	   like "remove@/kernel/dlm/gfs" */
-
-	if (strstr(buf, "kernel/dlm"))
-		return;
-
-	log_debug("uevent: %s", buf);
-
-	if (strstr(buf, "lock_module"))
-		lock_module = 1;
-
-	get_args(buf, &argc, argv, '/', 4);
-	if (argc != 4)
-		log_error("uevent message has %d args", argc);
-	act = argv[0];
-	sys = argv[2];
-
-	log_debug("kernel: %s %s", act, argv[3]);
-
-	if (!strcmp(act, "remove@")) {
+	if (!strcmp(uevent_vals[Env_ACTION], "remove")) {
 		/* We want to trigger the leave at the very end of the kernel's
 		   unmount process, i.e. at the end of put_super(), so we do the
 		   leave when the second uevent (from the gfs kobj) arrives. */
 
-		if (lock_module)
+		if (strcmp(uevent_vals[Env_SUBSYSTEM], "lock_dlm") == 0)
 			return;
+		if (group_mode == GROUP_LIBGROUP)
+			do_leave_old(fsname, 0);
+		else
+			do_leave(fsname, 0);
 
-		do_leave(argv[3], 0);
+	} else if (!strcmp(uevent_vals[Env_ACTION], "change")) {
+		int jid, status = -1, first = -1;
 
-	} else if (!strcmp(act, "change@")) {
-		if (!lock_module)
-			return;
+		if (!uevent_vals[Env_JID] ||
+		    (sscanf(uevent_vals[Env_JID], "%d", &jid) != 1))
+			jid = -1;
+
+		if (uevent_vals[Env_RECOVERY]) {
+			if (strcmp(uevent_vals[Env_RECOVERY], "Done") == 0)
+				status = LM_RD_SUCCESS;
+			if (strcmp(uevent_vals[Env_RECOVERY], "Failed") == 0)
+				status = LM_RD_GAVEUP;
+		}
+
+		if (uevent_vals[Env_FIRSTMOUNT] &&
+		    (strcmp(uevent_vals[Env_FIRSTMOUNT], "Done") == 0))
+			first = 1;
 
 		if (group_mode == GROUP_LIBGROUP)
-			process_recovery_uevent_old(argv[3]);
+			process_recovery_uevent_old(fsname, jid, status, first);
 		else
-			process_recovery_uevent(argv[3]);
-
-	} else if (!strcmp(act, "offline@")) {
-		if (!lock_module)
-			return;
-
-		if (group_mode == GROUP_LIBGROUP)
-			do_withdraw_old(argv[3]);
-		else
-			log_error("TODO withdraw for libcpg");
+			process_recovery_uevent(fsname, jid, status, first);
 
+	} else if (!strcmp(uevent_vals[Env_ACTION], "offline")) {
+		do_withdraw(fsname);
 	} else {
-		if (!lock_module)
-			return;
-
-		ping_kernel_mount(argv[3]);
+		ping_kernel_mount(fsname);
 	}
 }
 
@@ -567,19 +612,6 @@ static void query_mountgroup_nodes(int fd, char *name, int option, int max)
 		free(nodes);
 }
 
-void client_reply_remount(struct mountgroup *mg, int result)
-{
-	struct gfsc_mount_args *ma = &mg->mount_args;
-
-	log_group(mg, "client_reply_remount ci %d result %d",
-		  mg->remount_client, result);
-
-	do_reply(client[mg->remount_client].fd, GFSC_CMD_FS_REMOUNT,
-		 mg->name, result, ma, sizeof(struct gfsc_mount_args));
-
-	mg->remount_client = 0;
-}
-
 void client_reply_join(int ci, struct gfsc_mount_args *ma, int result)
 {
 	char *name = strstr(ma->table, ":") + 1;
@@ -613,7 +645,7 @@ void client_reply_join_full(struct mountgroup *mg, int result)
 	if (nodir_str[0])
 		strcat(mg->mount_args.hostdata, nodir_str);
  out:
-	log_group(mg, "client_reply_join_full ci %d result %d hostdata %s",
+	log_group(mg, "client_reply_join_full ci %d result %d %s",
 		  mg->mount_client, result, mg->mount_args.hostdata);
 
 	client_reply_join(mg->mount_client, &mg->mount_args, result);
@@ -743,16 +775,41 @@ static void do_join(int ci, struct gfsc_mount_args *ma)
 	client_reply_join(ci, ma, rv);
 }
 
-static void do_leave(char *table, int mnterr)
+/* The basic rule of withdraw is that we don't want to tell the kernel to drop
+   all locks until we know gfs has been stopped/blocked on all nodes.
+   A withdrawing node is very much like a readonly node, differences are
+   that others recover its journal when they remove it from the group,
+   and when it's been removed from the group, it tells the locally withdrawing
+   gfs to clear out locks. */
+
+static void do_withdraw(char *name)
 {
-	char *name = strstr(table, ":") + 1;
+	struct mountgroup *mg;
+	int rv;
 
-	log_debug("leave: %s mnterr %d", name, mnterr);
+	log_debug("withdraw: %s", name);
 
-	if (group_mode == GROUP_LIBGROUP)
-		gfs_leave_mountgroup_old(name, mnterr);
-	else
-		gfs_leave_mountgroup(name, mnterr);
+	if (!cfgd_enable_withdraw) {
+		log_error("withdraw feature not enabled");
+		return;
+	}
+
+	mg = find_mg(name);
+	if (!mg) {
+		log_error("do_withdraw no mountgroup %s", name);
+		return;
+	}
+
+	mg->withdraw_uevent = 1;
+
+	rv = run_dmsetup_suspend(mg, mg->mount_args.dev);
+	if (rv) {
+		log_error("do_withdraw %s: dmsetup %s error %d", mg->name,
+			  mg->mount_args.dev, rv);
+		return;
+	}
+
+	dmsetup_wait = 1;
 }
 
 static void do_mount_done(char *table, int result)
@@ -778,6 +835,53 @@ static void do_mount_done(char *table, int result)
 		gfs_mount_done(mg);
 }
 
+void client_reply_remount(struct mountgroup *mg, int ci, int result)
+{
+	do_reply(client[ci].fd, GFSC_CMD_FS_REMOUNT, mg->name, result,
+		 &mg->mount_args, sizeof(struct gfsc_mount_args));
+}
+
+/* mount.gfs creates a special ma->options string with only "ro" or "rw" */
+
+static void do_remount(int ci, struct gfsc_mount_args *ma)
+{
+	struct mountgroup *mg;
+	char *name = strstr(ma->table, ":") + 1;
+	int ro = 0, result = 0;
+
+	log_debug("remount: %s ci %d options %s", name, ci, ma->options);
+
+	mg = find_mg(name);
+	if (!mg) {
+		log_error("remount: %s not found", name);
+		result = -1;
+		goto out;
+	}
+
+	if (mg->spectator) {
+		log_error("remount of spectator not allowed");
+		result = -1;
+		goto out;
+	}
+
+	if (!strcmp(ma->options, "ro"))
+		ro = 1;
+
+	if ((mg->ro && ro) || (!mg->ro && !ro))
+		goto out;
+
+	if (group_mode == GROUP_LIBGROUP) {
+		/* the receive calls client_reply_remount */
+		mg->remount_client = ci;
+		send_remount_old(mg, ma);
+		return;
+	}
+
+	send_remount(mg, ma);
+ out:
+	client_reply_remount(mg, ci, result);
+}
+
 void process_connection(int ci)
 {
 	struct gfsc_header h;
@@ -839,7 +943,10 @@ void process_connection(int ci)
 		break;
 
 	case GFSC_CMD_FS_LEAVE:
-		do_leave(ma->table, h.data);
+		if (group_mode == GROUP_LIBGROUP)
+			do_leave_old(ma->table, h.data);
+		else
+			do_leave(ma->table, h.data);
 		break;
 
 	case GFSC_CMD_FS_MOUNT_DONE:
@@ -847,13 +954,7 @@ void process_connection(int ci)
 		break;
 
 	case GFSC_CMD_FS_REMOUNT:
-		if (group_mode == GROUP_LIBGROUP)
-			remount_mountgroup_old(ci, ma);
-#if 0
-		/* FIXME */
-		else
-			remount_mountgroup(ci, ma);
-#endif
+		do_remount(ci, ma);
 		break;
 
 	default:
@@ -883,7 +984,7 @@ static void process_listener(int ci)
 	log_debug("client connection %d fd %d", i, fd);
 }
 
-static int setup_listener(char *sock_path)
+static int setup_listener(const char *sock_path)
 {
 	struct sockaddr_un addr;
 	socklen_t addrlen;
@@ -935,11 +1036,18 @@ void query_unlock(void)
 static void *process_queries(void *arg)
 {
 	struct gfsc_header h;
-	int s = *((int *)arg);
-	int f, rv;
+	int f, rv, s;
+
+	rv = setup_listener(GFSC_QUERY_SOCK_PATH);
+	if (rv < 0)
+		return NULL;
+
+	s = rv;
 
 	for (;;) {
 		f = accept(s, NULL, NULL);
+		if (f < 0)
+			return NULL;
 
 		rv = do_read(f, &h, sizeof(h));
 		if (rv < 0) {
@@ -987,19 +1095,13 @@ static void *process_queries(void *arg)
 
 static int setup_queries(void)
 {
-	int rv, s;
-
-	rv = setup_listener(GFSC_QUERY_SOCK_PATH);
-	if (rv < 0)
-		return rv;
-	s = rv;
+	int rv;
 
 	pthread_mutex_init(&query_mutex, NULL);
 
-	rv = pthread_create(&query_thread, NULL, process_queries, &s);
+	rv = pthread_create(&query_thread, NULL, process_queries, NULL);
 	if (rv < 0) {
 		log_error("can't create query thread");
-		close(s);
 		return rv;
 	}
 	return 0;
@@ -1007,15 +1109,10 @@ static int setup_queries(void)
 
 void cluster_dead(int ci)
 {
-	log_error("cluster is down, exiting");
+	if (!cluster_down)
+		log_error("cluster is down, exiting");
 	daemon_quit = 1;
-}
-
-static void dlmcontrol_dead(int ci)
-{
-	if (!list_empty(&mountgroups))
-		log_error("dlm_controld is gone");
-	client_dead(ci);
+	cluster_down = 1;
 }
 
 static void loop(void)
@@ -1062,9 +1159,21 @@ static void loop(void)
 			goto out;
 		client_add(rv, process_groupd, cluster_dead);
 
-		group_mode = GROUP_LIBGROUP;
-		if (cfgd_groupd_compat == 2)
-			set_group_mode();
+		switch (cfgd_groupd_compat) {
+		case 1:
+			group_mode = GROUP_LIBGROUP;
+			rv = 0;
+			break;
+		case 2:
+			rv = set_group_mode();
+			break;
+		default:
+			log_error("inval groupd_compat %d", cfgd_groupd_compat);
+			rv = -1;
+			break;
+		}
+		if (rv < 0)
+			goto out;
 	}
 	log_debug("group_mode %d compat %d", group_mode, cfgd_groupd_compat);
 
@@ -1076,16 +1185,26 @@ static void loop(void)
 		 */
 
 		rv = setup_cpg();
+		if (rv < 0)
+			goto out;
+		client_add(rv, process_cpg, cluster_dead);
+
+		rv = set_protocol();
 		if (rv < 0)
 			goto out;
 
 		rv = setup_dlmcontrol();
 		if (rv < 0)
 			goto out;
-		client_add(rv, process_dlmcontrol, dlmcontrol_dead);
+		client_add(rv, process_dlmcontrol, cluster_dead);
 
 	} else if (group_mode == GROUP_LIBGROUP) {
 
+		/* the stable2 default is zero plock_ownership which we need
+		   to match or protocol_active will differ */
+		if (using_default_plock_ownership)
+			cfgd_plock_ownership = 0;
+
 		/*
 		 * The old, bad, way of doing things using libgroup.
 		 * code in: cpg-old.c group.c plock.c
@@ -1096,6 +1215,10 @@ static void loop(void)
 			goto out;
 		client_add(rv, process_cpg_old, cluster_dead);
 
+		rv = setup_misc_devices();
+		if (rv < 0)
+			goto out;
+
 		rv = setup_plocks();
 		if (rv < 0)
 			goto out;
@@ -1116,7 +1239,6 @@ static void loop(void)
 			goto out;
 		}
 
-		/* FIXME: lock/unlock around operations that take a while */
 		query_lock();
 
 		for (i = 0; i <= client_maxi; i++) {
@@ -1131,10 +1253,13 @@ static void loop(void)
 				deadfn(i);
 			}
 		}
+		query_unlock();
 
 		if (daemon_quit)
 			break;
 
+		query_lock();
+
 		poll_timeout = -1;
 
 		if (poll_dlm) {
@@ -1162,10 +1287,15 @@ static void loop(void)
 					poll_timeout = -1;
 			}
 		}
-
 		query_unlock();
 	}
  out:
+	if (group_mode == GROUP_LIBCPG)
+		close_cpg();
+	else if (group_mode == GROUP_LIBGROUP) {
+		close_plocks();
+		close_cpg_old();
+	}
 	if (cfgd_groupd_compat)
 		close_groupd();
 	close_logging();
@@ -1226,8 +1356,8 @@ static void print_usage(void)
 	printf("\n");
 	printf("Options:\n");
 	printf("\n");
-	printf("  -D           Enable debugging code and don't fork\n");
-	printf("  -L <num>     Enable (1) or disable (0) debugging to logsys (default %d)\n", DEFAULT_DEBUG_LOGSYS);
+	printf("  -D           Enable debugging to stderr and don't fork\n");
+	printf("  -L           Enable debugging to log file\n");
 	printf("  -g <num>     groupd compatibility mode, 0 off, 1 on, 2 detect\n");
 	printf("               0: use libcpg, no backward compat, best performance\n");
 	printf("               1: use libgroup for compat with cluster2/rhel5\n");
@@ -1254,7 +1384,7 @@ static void print_usage(void)
 	printf("  -V           Print program version information, then exit\n");
 }
 
-#define OPTION_STRING "L:DKg:w:f:q:d:p:Pl:o:t:c:a:hV"
+#define OPTION_STRING "LDKg:w:f:q:d:p:Pl:o:t:c:a:hV"
 
 static void read_arguments(int argc, char **argv)
 {
@@ -1271,8 +1401,8 @@ static void read_arguments(int argc, char **argv)
 			break;
 
 		case 'L':
-			optd_debug_logsys = 1;
-			cfgd_debug_logsys = atoi(optarg);
+			optd_debug_logfile = 1;
+			cfgd_debug_logfile = 1;
 			break;
 
 		case 'g':
@@ -1349,9 +1479,9 @@ static void read_arguments(int argc, char **argv)
 		};
 	}
 
-	if (!optd_debug_logsys && getenv("GFS_CONTROLD_DEBUG")) {
-		optd_debug_logsys = 1;
-		cfgd_debug_logsys = atoi(getenv("GFS_CONTROLD_DEBUG"));
+	if (getenv("GFS_CONTROLD_DEBUG")) {
+		optd_debug_logfile = 1;
+		cfgd_debug_logfile = 1;
 	}
 }
 
@@ -1388,26 +1518,26 @@ static void set_scheduler(void)
 int main(int argc, char **argv)
 {
 	INIT_LIST_HEAD(&mountgroups);
-
-	init_logging();
+	INIT_LIST_HEAD(&withdrawn_mounts);
 
 	read_arguments(argc, argv);
 
-	lockfile();
-
 	if (!daemon_debug_opt) {
 		if (daemon(0, 0) < 0) {
 			perror("daemon error");
 			exit(EXIT_FAILURE);
 		}
 	}
+	lockfile();
+	init_logging();
+	log_level(LOG_INFO, "gfs_controld %s started", RELEASE_VERSION);
 	signal(SIGTERM, sigterm_handler);
-
 	set_scheduler();
 	set_oom_adj(-16);
 
 	loop();
 
+	unlink(LOCKFILE_NAME);
 	return 0;
 }
 
@@ -1429,6 +1559,7 @@ void daemon_dump_save(void)
 
 int daemon_debug_opt;
 int daemon_quit;
+int cluster_down;
 int poll_ignore_plock;
 int poll_dlm;
 int plock_fd;
@@ -1444,7 +1575,11 @@ int dump_wrap;
 char plock_dump_buf[GFSC_DUMP_SIZE];
 int plock_dump_len;
 int dmsetup_wait;
-cpg_handle_t libcpg_handle;
+cpg_handle_t cpg_handle_daemon;
 int libcpg_flow_control_on;
 int group_mode;
+uint32_t plock_minor;
+uint32_t old_plock_minor;
+struct list_head withdrawn_mounts;
+int using_default_plock_ownership;
 
